"""Shared fixture set."""

import os
import pathlib
import shutil
import tempfile
from typing import Union

import pytest
from _pytest.config import ExitCode
from _pytest.fixtures import SubRequest
from _pytest.main import Session

import dagsesh.lazy
from dagsesh.logging_config import log


@pytest.fixture
def working_dir(request: "SubRequest") -> str:
    """Temporary working directory."""

    def fin() -> None:
        """Tear down."""
        log.info('Deleting temporary test directory: "%s"', dirpath)
        shutil.rmtree(dirpath)

    request.addfinalizer(fin)
    dirpath = tempfile.mkdtemp()
    log.info('Created temporary test directory: "%s"', dirpath)

    return dirpath


@pytest.fixture(scope="session")
def dagbag():
    """Set up the Airflow DagBag common to pytest.Session."""
    af_models = dagsesh.lazy.Loader("models", globals(), "airflow.models")

    return af_models.DagBag()


def pytest_sessionstart(session: Session) -> None:  # pylint: disable=unused-argument
    """Set up the Airflow context with appropriate config for test."""
    airflow_home = tempfile.mkdtemp()
    os.environ["AIRFLOW_HOME"] = airflow_home
    log.info('Temporary Airflow home (AIRFLOW_HOME): "%s"', airflow_home)

    project_dir = os.environ.get("PROJECT_SOURCE_DIR")
    if not project_dir:
        cwd = pathlib.Path.cwd()
        project_dir = os.path.join(cwd, cwd.resolve().name.lower().replace("-", "_"))

    os.environ["AIRFLOW__CORE__UNIT_TEST_MODE"] = "true"
    os.environ["AIRFLOW__CORE__DAGS_FOLDER"] = os.path.join(project_dir, "dags")
    os.environ["AIRFLOW__CORE__PLUGINS_FOLDER"] = os.path.join(project_dir, "plugins")
    os.environ["AIRFLOW__CORE__LOAD_EXAMPLES"] = "false"
    os.environ["AIRFLOW__CORE__FERNET_KEY"] = (
        "LFKF4PSrAOG-kbxOouoLj8Du2QCnsp9qw7G21-WPsLU="
    )
    db_url = f"sqlite:///{airflow_home}/airflow.db;Version=3;Journal Mode=Off;"
    os.environ["AIRFLOW__DATABASE__LOAD_DEFAULT_CONNECTIONS"] = "false"
    os.environ["AIRFLOW__DATABASE__SQL_ALCHEMY_CONN"] = db_url

    log.info('Airflow project directory: "%s"', project_dir)
    for key, val in [x for x in os.environ.items() if x[0].startswith("AIRFLOW__")]:
        log.info("Environment Variable %s: %s", key, val)

    if os.environ.get("AIRFLOW__DAGSESH__PRIME_TEST_CONTEXT") == "true":
        log.info("AIRFLOW__DAGSESH__PRIME_TEST_CONTEXT is set")
        utils_db = dagsesh.lazy.Loader("utils_db", globals(), "airflow.utils.db")
        utils_db.upgradedb()


def pytest_sessionfinish(  # pylint: disable=unused-argument
    session: Session,
    exitstatus: Union[int, ExitCode],
) -> None:
    """Tear down the Airflow context."""
    log.info(
        'Deleting working temporary test directory: "%s"', os.environ["AIRFLOW_HOME"]
    )
    shutil.rmtree(os.environ["AIRFLOW_HOME"])



"""Delay imports of Python modules and packages."""

import importlib
import types


class Loader(types.ModuleType):
    """Lazily import a module."""

    def __init__(self, local_name: str, parent_module_globals: dict, module_name: str):
        """Initialise a lazy importer for a module.

        This class enables deferred loading of a module until its first use,
        improving startup performance by avoiding immediate imports.

        Parameters:
            local_name: The name under which the imported module will be accessible
                within the local namespace.
            parent_module_globals: The globals dictionary of the parent module where
                the lazy import will be registered.
            module_name: The name of the module to be lazily imported.

        """
        self.__local_name = local_name
        self.__parent_module_globals = parent_module_globals

        super().__init__(module_name)

    @property
    def local_name(self) -> str:
        """Get the local_name."""
        return self.__local_name

    @property
    def parent_module_globals(self) -> dict:
        """Get the parent_module_globals."""
        return self.__parent_module_globals

    def _load(self) -> types.ModuleType:
        """Import the target module and insert it into the parent's namespace.

        The method will update this object's dict so that if someone keeps a reference to the
        LazyLoader, lookups are efficient (__getattr__ is only called on lookups that fail).

        Returns:
            Python module.

        """
        module = importlib.import_module(self.__name__)
        self.parent_module_globals[self.local_name] = module

        self.__dict__.update(module.__dict__)

        return module

    def __getattr__(self, item: str) -> str:
        """Delegate attribute access to the lazily loaded module.

        This method is called when an attribute not found in the instance is accessed.
        It triggers the lazy loading of the module and then attempts to retrieve
        the requested attribute from it.

        Parameters:
            item: The name of the attribute to retrieve from the loaded module.

        Returns:
            The value of the requested attribute from the underlying module.

        Raises:
            AttributeError: If the attribute does not exist in the loaded module.

        """
        module = self._load()

        return getattr(module, item)



"""Logging configuration."""

from __future__ import annotations

import logging.config
import os
from typing import TYPE_CHECKING

import structlog

if TYPE_CHECKING:
    import structlog.stdlib


timestamper = structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S")
pre_chain = [
    # Add the log level and a timestamp to the event_dict if the log entry
    # is not from structlog.
    structlog.stdlib.add_log_level,
    # Add extra attributes of LogRecord objects to the event dictionary
    # so that values passed in the extra parameter of log methods pass
    # through to log output.
    structlog.stdlib.ExtraAdder(),
    timestamper,
]


def extract_from_record(
    _: structlog.stdlib._FixedFindCallerLogger, __: str, event_dict: dict
) -> dict:
    """Extract thread and process names and add them to the event dict."""
    record = event_dict["_record"]
    event_dict["thread"] = record.threadName

    return event_dict


logging.config.dictConfig(
    {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "plain": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processors": [
                    structlog.stdlib.ProcessorFormatter.remove_processors_meta,
                    structlog.dev.ConsoleRenderer(colors=False),
                ],
                "foreign_pre_chain": pre_chain,
            },
            "colored": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processors": [
                    extract_from_record,
                    structlog.stdlib.ProcessorFormatter.remove_processors_meta,
                    structlog.dev.ConsoleRenderer(colors=True),
                ],
                "foreign_pre_chain": pre_chain,
            },
        },
        "handlers": {
            "default": {
                "level": os.environ.get("LOG_LEVEL", "INFO").upper(),
                "class": "logging.StreamHandler",
                "formatter": "plain",
            },
        },
        "loggers": {
            "": {
                "handlers": ["default"],
                "level": os.environ.get("LOG_LEVEL", "INFO").upper(),
                "propagate": True,
            },
        },
    }
)
structlog.configure(
    processors=[
        structlog.stdlib.add_log_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        timestamper,
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
    ],
    logger_factory=structlog.stdlib.LoggerFactory(),
    wrapper_class=structlog.stdlib.BoundLogger,
    cache_logger_on_first_use=True,
)

log = structlog.get_logger("webapp")
